/*	scaleswg.c 		

	Provide statistical estimates using a Gaussian distribution

	copyright (c) 1995, 1996 William R. Pearson

	This code provides multiple methods for scaling sequence
	similarity scores to correct for length effects.

	Currently, three methods are available:

	pst.zsflag = 0 - no scaling
	pst.zsflag = 1 - regression-scaled scores
	pst.zsflag = 2 - log-scaled scores
	pst.zsflag = 3 - scaling using Altschul's parameters
	pst.zsflag = 4 - regression-scaled with iterative outlier removal
	pst.zsflag = 5 = like 1, but length scaled variance
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#if!defined(__MWERKS__) && !defined(MacOSX)
#include <values.h>
#else
#include <limits.h>	/* needed for LONG_MAX */
#define MAXLONG LONG_MAX
#endif

#ifndef M_SQRT2
#define M_SQRT2 1.41421356237
#endif
#define LN2 0.69314718055994530942

#include "defs.h"

#define MAXHIST 50
#define MAX_LLEN 200
#define LHISTC 5
#define VHISTC 5
#define MAX_SSCORE 300

FILE *tfd;

/* extern */ int histflg;
/* extern */ int *hist, histint, min_hist, max_hist, maxh;
double ks_dev;
int ks_df;
long num_db_entries, hist_db_entries, z_calls;
static int n_trimmed, n1_trimmed, nb_trimmed, nb_tot;
static int fit_flag;

struct total_db_str {
  unsigned long length;
  int carry;
} total_db;

#define LENGTH_CUTOFF 10 /* minimum database sequence length allowed, for fitting */

#define LN_FACT 10.0

#define LN200 5.2983173666
#define ZS_MAX 50.0

int *llen_hist=NULL, *score_sums, *score2_sums;
static int *h_len;
double *score_var, *score_ave, **s_idx;
static double logK, lambda, rho, rho_e, mu, mu_e;
static double rho2, mu2, mean_var, var_e, mean_length;
static int max_llen, min_llen, max_sscore;
static int max_length, min_length, zero_scores;
static double avg_entry_length, var_cutoff;

#include "param.h"
#ifndef PCOMPLIB
#include "mw.h"
#else
#include "p_mw.h"
#endif

static char f_string[MAX_STR];

float find_zr(int score, int len, double comp);
float find_zr2(int score, int len, double comp);
float find_z(int score, int len, double comp);
float find_zn(int score, int len, double comp);
float find_za(int score, int len, double comp);

float (*find_zp)(int score, int len, double comp) = &find_zr;

void proc_hist_r(struct stat_str *sptr, int n, struct pstruct pst);
void proc_hist_r2(struct stat_str *sptr, int n, struct pstruct pst);
void proc_hist_n(struct stat_str *sptr, int n, struct pstruct pst);
void proc_hist_a(struct stat_str *sptr, int n, struct pstruct pst);

void s_sort (double **ptr, int nbest);
void st_sort ( int *sptr, int n);
void inithist();
void free_hist();
void inithistz(int);
void addhist(double (*func)(), int score, int length);
void addhistz(float zs);
void addhistzp(float zs);
void fit_llen();
void fit_llen2();
void fit_llens();
void prune_hist(double (*func)(), int score, int length);
void set_db_size(), update_db_size();

void
process_hist(struct stat_str *sptr, int nstats, struct pstruct pst)
{
  if (pst.zsflag==2) {
    find_zp = &find_zr;
    proc_hist_r(sptr, nstats,pst);
  }
  else if (pst.zsflag==1) {
    find_zp = &find_zr;
    proc_hist_r(sptr, nstats,pst);
  }
  else if (pst.zsflag==3) {
    find_zp = &find_za;
    proc_hist_a(sptr, nstats,pst);
  }
  else if (pst.zsflag==4) {
    find_zp = &find_zr;
    proc_hist_r2(sptr,nstats,pst);
  }
  else if (pst.zsflag==5) {
    find_zp = &find_zr2;
    proc_hist_r(sptr,nstats,pst);
  }
  else {
    find_zp = &find_zn;
    proc_hist_n(sptr,nstats,pst);
  }
}

double eq_s(int s, int n)
{
  return (double)s;
}
     

void
proc_hist_r(struct stat_str *sptr, int nstats, struct pstruct pst)
{
  int i, j, llen, hscore;
  double zs;
  double eq_s();

  fit_flag=1;

  inithist();

  for (i = 0; i<nstats; i++)
    addhist(&eq_s,sptr[i].score,sptr[i].n1);
  hist_db_entries = num_db_entries;

  fit_llen();	/* now we have rho, mu, rho2, mu2, mean_var to set
		   the parameters for the histogram */

  if (!fit_flag) {
    find_zp = &find_zn;
    proc_hist_n(sptr,nstats,pst);
    return;
  }

  n_trimmed= n1_trimmed = nb_trimmed = 0;

  if (fit_flag) {
    for (i = 0; i < nstats; i++) {
      zs = find_zr(sptr[i].score,sptr[i].n1,sptr[i].comp);
      if (zs < 20.0 || zs > 100.0 ) {
	prune_hist(&eq_s,sptr[i].score,sptr[i].n1);
      }
    }
  }

  /*  fprintf(stderr,"Z-trimmed %d entries with z > 5.0\n", n_trimmed); */

  if (fit_flag) fit_llens();

  /*   fprintf(stderr,"Bin-trimmed %d entries in %d bins\n", n1_trimmed,nb_trimmed); */

  /* put all the scores in the histogram */

  hist_db_entries = num_db_entries;
  inithistz(MAXHIST);
  for (i = 0; i < nstats; i++)
    addhistzp(find_zr(sptr[i].score,sptr[i].n1,sptr[i].comp));

  if (pst.zsflag == 5) 
    sprintf(f_string,"Expectation_v fit: rho(ln(x))= %6.4lf+/-%6.3lg; mu= %6.4lf+/-%6.3lf;\n rho2=%6.2lf; mu2= %6.2lf, Z-trim: %d  B-trim: %d in %d/%d",
	  rho*LN_FACT,sqrt(rho_e),mu,sqrt(mu_e),rho2,mu2,
	    n_trimmed, n1_trimmed, nb_trimmed, nb_tot);
  else 
    sprintf(f_string,"Expectation_n fit: rho(ln(x))= %6.4lf+/-%6.3lg; mu= %6.4lf+/-%6.3lf;\n mean_var=%6.4lf+/-%6.3lf, Z-trim: %d  B-trim: %d in %d/%d",
	  rho*LN_FACT,sqrt(rho_e),mu,sqrt(mu_e),mean_var,sqrt(var_e),
	    n_trimmed, n1_trimmed, nb_trimmed, nb_tot);
}

void
proc_hist_r2(struct stat_str *sptr, int nstats, struct pstruct pst)
{
  int i, j, llen, hscore, nit, nprune;
  double zs;
  double eq_s();

  fit_flag=1;

  inithist();

  for (i = 0; i<nstats; i++)
    addhist(&eq_s,sptr[i].score,sptr[i].n1);
  hist_db_entries = num_db_entries;

  n_trimmed= n1_trimmed = nb_trimmed = nit = 0;
  while (nit++ < 5) {
    nprune = 0;
    fit_llen2();

    for (i = 0; i < nstats; i++) {
      if (sptr[i].n1 < 0) continue;
      zs = find_zr(sptr[i].score,sptr[i].n1,sptr[i].comp);
      if (zs < 20.0 || zs > 120.0 ) {
	nprune++;
	prune_hist(&eq_s,sptr[i].score,sptr[i].n1);
	sptr[i].n1 = -sptr[i].n1;
      }
    }
    /*    fprintf(stderr," %d Z-trimmed at %d\n",nprune,nit); */
    if (nprune < LHISTC) {
      break;
    }
  }
  fit_llen();

  hist_db_entries = num_db_entries;
  inithistz(MAXHIST);
  for (i = 0; i < nstats; i++) {
    if (sptr[i].n1 < 0) sptr[i].n1 = -sptr[i].n1;
    addhistzp(find_zr(sptr[i].score,sptr[i].n1,sptr[i].comp));
  }

  sprintf(f_string,"Expectation_i fit: rho(ln(x))= %6.4lf+/-%6.3lg; mu= %6.4lf+/-%6.3lf;\n mean_var=%6.4lf+/-%6.3lf Z-trim: %d N-it: %d",
	  rho*LN_FACT,sqrt(rho_e),mu,sqrt(mu_e),mean_var,sqrt(var_e),
	  n_trimmed, nit);
}


/* this procedure implements Altschul's pre-calculated values for lambda, K */


#include "alt_parms.h"

static double K, Lambda, H, a_n0f, a_n0;

void
proc_hist_a(struct stat_str *sptr, int nstats, struct pstruct pst)
{
  int i, j, llen, hscore, r_v;
  double s_score, s2_score, ln_std, l_score;

  double zs;

  if (strcmp(pst.pamfile,"BL50")==0)
      r_v = look_p(bl50_p,pst.gdelval,pst.ggapval,&K,&Lambda,&H);
  else if (strcmp(pst.pamfile,"BL62")==0)
      r_v = look_p(bl62_p,pst.gdelval,pst.ggapval,&K,&Lambda,&H);
  else if (strcmp(pst.pamfile,"250")==0)
      r_v = look_p(p250_p,pst.gdelval,pst.ggapval,&K,&Lambda,&H);
  else r_v = 0;

  if (r_v == 0) {
    fprintf(stderr,"Parameters not available for: %s: %d/%d\n",
	    pst.pamfile,pst.gdelval,pst.ggapval);

    find_zp = find_zr;
    proc_hist_r(sptr, nstats,pst);
    return;
    }

  /*
    fprintf(stderr," the parameters are: Lambda: %5.3f K: %5.3f H: %5.3f\n",
	    Lambda, K, H);
	    */

    a_n0 = (double)pst.n0;
    a_n0f = log (K * a_n0)/H;

    num_db_entries = total_db.length=total_db.carry = 0;

    inithistz(MAXHIST);
    for (i = 0; i < nstats; i++) {
      zs = find_za(sptr[i].score, sptr[i].n1,sptr[i].comp);
      addhistz(zs);
    }

    sprintf(f_string,"Altschul/Gish params: n0: %d Lambda: %5.3f K: %5.3f H: %5.3f",
	    pst.n0,Lambda, K, H);
}

look_p(struct alt_p parm[], int gap, int ext,
       double *K, double *Lambda, double *H)
{
  int i;

  gap = -gap;
  ext = -ext;

  if (gap > parm[1].gap) {
    *K = parm[0].K;
    *Lambda = parm[0].Lambda;
    *H = parm[0].H;
    return 1;
  }

  for (i=1; parm[i].gap > 0; i++) {
    if (parm[i].gap > gap) continue;
    else if (parm[i].gap == gap && parm[i].ext > ext ) continue;
    else if (parm[i].gap == gap && parm[i].ext == ext) {
      *K = parm[i].K;
      *Lambda = parm[i].Lambda;
      *H = parm[i].H;
      return 1;
    }
    else break;
  }
  return 0;
}

void
proc_hist_n(struct stat_str *sptr, int nstats, struct pstruct pst)
{
  int i, j, llen, hscore;
  double s_score, s2_score, ssd;
  int n_hist, n_size, nit;
  int *ss;

  double zs;

  num_db_entries = total_db.length = total_db.carry = 0;

  if ((ss=calloc(nstats,sizeof(int)))==NULL) {
    fprintf(stderr," cannot allocate ss matrix\n");
    return ;
  }

  for (i=0; i<nstats; i++)
    ss[i] = eq_s(sptr[i].score,sptr[i].n1);

  st_sort(ss,nstats);

  s_score = s2_score = 0.0;

  for ( j = 0, i = 0; i < nstats; i++) {
    if (ss[i] > 0) {
      s_score += (ssd=(double)ss[i]);
      s2_score += ssd * ssd;
      j++;
    }
  }

  if (j > 1 ) {
    mu = s_score/(double)j;
    mean_var = s2_score - (double)j * mu * mu;
    mean_var /= (double)(j-1);
  }
  else {
    mu = 50.0;
    mean_var = 10.0;
  }

  
  /* now remove some scores */

  nit = 5;
  while (nit-- > 0) {
    n_trimmed = 0;

    for (i=0; i< nstats; i++) {
      if (sptr[i].n1 < 0) continue;
      ssd = find_zn(sptr[i].score,sptr[i].n1,sptr[i].comp);
      if (ssd > 120.0 || ssd < 20.0) {
	/*      fprintf(stderr,"removing %3d %3d %4.1f\n",
		sptr[i].score, sptr[i].n1,ssd); */
	ssd = eq_s(sptr[i].score,sptr[i].n1);
	s_score -= ssd;
	s2_score -= ssd*ssd;
	j--;
	n_trimmed++;
	sptr[i].n1 = -sptr[i].n1;
      }
    }

    if (j > 1 ) {
      mu = s_score/(double)j;
      mean_var = s2_score - (double)j * mu * mu;
      mean_var /= (double)(j-1);
    }
    else {
      mu = 50.0;
      mean_var = 10.0;
    }

    if (n_trimmed < LHISTC) {
      /*
	fprintf(stderr,"nprune %d at %d\n",nprune,nit);
	*/
      break;
    }
  }

  free(ss);

  inithistz(MAXHIST);
  for (i = 0; i < nstats; i++) {
    if (sptr[i].n1 < 0) sptr[i].n1 = -sptr[i].n1;
    addhistz(find_zn(sptr[i].score,sptr[i].n1,sptr[i].comp));
  }

  sprintf(f_string,"unscaled statistics: mu= %6.4lf  var=%6.4lf",mu,mean_var);
}

void
alloc_hist()
{
  int i;

  if (llen_hist == NULL) {
    llen_hist = (int *)calloc((size_t)(max_llen+1),sizeof(int));
    score_sums = (int *)calloc((size_t)(max_llen + 1),sizeof(int));
    score2_sums = (int *)calloc((size_t)(max_llen + 1),sizeof(int));
    score_ave = (double *)calloc((size_t)(max_llen + 1),sizeof(double));
    score_var = (double *)calloc((size_t)(max_llen + 1),sizeof(double));
    s_idx = (double **)calloc((size_t)(max_llen+2),sizeof(double *));
  }
  else {
    for (i=0; i< max_llen+1; i++) {
      llen_hist[i] = score_sums[i] = score2_sums[i] = 0;
      score_ave[i] = score_var[i] = 0.0;
    }
  }
}
  
void
free_hist()
{
  if (llen_hist!=NULL) {
    free(s_idx);
    free(score_var);
    free(score_ave);
    free(score2_sums);
    free(score_sums);
    free(llen_hist);
    llen_hist=NULL;
  }
}

void
inithist()
{
  int i;

  max_llen = MAX_LLEN;
  max_sscore = MAX_SSCORE;

  alloc_hist();

  zero_scores = 0;
  min_length = 10000;
  max_length = 0;
  num_db_entries = 0;
  total_db.length = total_db.carry = 0;

}

void
addhist(double (*func)(), int score, int length)
{
  int llen; 
  double f_score;

  if ( score<=0 || length < LENGTH_CUTOFF) {
    zero_scores++;
    return;
  }

  if (length > max_length) max_length = length;
  if (length < min_length) min_length = length;
  if (score > max_sscore) score = max_sscore;

  llen = (int)(LN_FACT*log((double)length)+0.5);

  if (llen < 0 ) llen = 0;
  if (llen > max_llen) llen = max_llen;
  llen_hist[llen]++;
  f_score = (*func)(score, length);
  score_sums[llen] += f_score;
  score2_sums[llen] += f_score * f_score;

  num_db_entries++;
  total_db.length += length;
  if (total_db.length > MAXLONG) {
    total_db.carry++;
    total_db.length -= MAXLONG;
  }
}

/* histogram will go from z-scores of 20 .. 100 with mean 50 and z=10 */

void
inithistz(mh)
     int mh;
{
  min_hist = 20;
  max_hist = 120;
  z_calls = 0;

  histint = (int)((double)(max_hist - min_hist + 2)/(double)mh+0.5);
  maxh = (int)((double)(max_hist - min_hist + 2)/(double)histint+0.5) ;

  if ((hist=(int *)calloc((size_t)maxh,sizeof(int)))==NULL) {
    fprintf(stderr," cannot allocate %d for histogram\n",maxh);
    histflg = 0;
  }
}

void
addhistz(float zs)
{
  int ih, zi;
  int inv;
  static int iset=0;

  z_calls++;


  zi = (int)(zs + 0.5);

  if ((zi >= 0) && (zi <= 120)) hist_db_entries = ++num_db_entries;

  if (zi < min_hist) zi = min_hist;
  if (zi > max_hist) zi = max_hist;
  
  ih = (zi - min_hist)/histint;

  hist[ih]++;
}

/* addhistzp() does not increase num_db_entries since addhist did it already */
void
addhistzp(float zs)
{
  int ih, zi;
  int inv;
  static int iset=0;


  z_calls++;

  zi = (int)(zs + 0.5);

  if (zi < min_hist) zi = min_hist;
  if (zi > max_hist) zi = max_hist;
  
  ih = (zi - min_hist)/histint;

  hist[ih]++;
}

void
prune_hist(double (*func)(), int score, int length)
{
  int llen;
  double f_score;

  if (score <= 0 || length < LENGTH_CUTOFF) return;

  if (score > max_sscore) score = max_sscore;

  llen = (int)(LN_FACT*log((double)length)+0.5);

  if (llen < 0 ) llen = 0;
  if (llen > max_llen) llen = max_llen;
  llen_hist[llen]--;
  f_score = (*func)(score,length);
  score_sums[llen] -= f_score;
  score2_sums[llen] -= f_score * f_score;

  num_db_entries--;
  n_trimmed++;
  if (length < total_db.length) {
    total_db.length -= length;
  }
  else {
    total_db.carry--;
    total_db.length += (MAXLONG - (unsigned long)length);
  }
}  
 
void
update_db_size(int nlib)
{
  num_db_entries = nlib - n_trimmed;
}

void
set_db_size(int nlib)
{
  num_db_entries = nlib;
}

void
fit_llen()
{
  int j;
  int n, s, s2, ej, last_ej, cell_size;
  int n_size;
  double mean, x, y, y2, t, t2, u, u2, v, p, z;
  double mean_x, mean_y, var_x, var_y, covar_xy;
  double mean_y2, covar_xy2, var_y2, dllj;
  double ex, exl;

  double sum_x, sum_y, sum_x2, sum_xy, sum_v, det, n_w;
  
/* now fit scores to best linear function of log(n), using
   simple linear regression */
  
  for (min_llen=0; min_llen < max_llen; min_llen++)
    if (llen_hist[min_llen]) break;
  min_llen--;

  for (n_size=0,j = min_llen; j < max_llen; j++) {
    if (llen_hist[j] > 1) {
      dllj = (double)llen_hist[j];
      score_var[j] = (double)score2_sums[j]/dllj
	- ((double)score_sums[j]/dllj)
	* ((double)score_sums[j]/dllj);
      score_var[j] /= (double)(llen_hist[j]-1);
      if (score_var[j] <= 1.0 ) score_var[j] = 1.0;
      n_size++;
    }
  }
	  
  nb_tot = n_size;

  n_w = 0.0;
  sum_x = sum_y = sum_x2 = sum_xy = sum_v = 0;
  for (j = min_llen; j < max_llen; j++)
    if (llen_hist[j] > 1) {
      x = j + 0.5;
      dllj = (double)llen_hist[j];
      n_w += dllj/score_var[j];
      sum_x +=   dllj * x / score_var[j] ;
      sum_y += (double)score_sums[j] / score_var[j];
      sum_x2 +=  dllj * x * x /score_var[j];
      sum_xy +=  x * (double)score_sums[j]/score_var[j];
    }

  if (n_w < 0.001) {
    fit_flag = 0;
    rho = 0;
    return;
  }

  if (n_size < 5 ) {
    fit_flag=0;
    rho = 0;
    mu = sum_y/n_w;
    return;
  }
  else {
    det = n_w * sum_x2 - sum_x * sum_x;
    if (det > 0.001) {
      rho = (n_w * sum_xy  - sum_x * sum_y)/det;
      rho_e = n_w/det;
      mu = (sum_x2 * sum_y - sum_x * sum_xy)/det;
      mu_e = sum_x2/det;
    }
    else {
      fit_flag = 0;
      rho = 0;
      mu = sum_y/n_w;
      return;
    }
  }

  det = n_w * sum_x2 - sum_x * sum_x;
  rho = (n_w * sum_xy  - sum_x * sum_y)/det;
  mu = (sum_x2 * sum_y - sum_x * sum_xy)/det;

  n = 0;
  mean_x = mean_y = mean_y2 = 0.0;
  var_x = var_y = 0.0;
  covar_xy = covar_xy2 = 0.0;

  for (j = min_llen; j <= max_llen; j++) 
   if (llen_hist[j] > 1 ) {
      n += llen_hist[j];
      x = (double)j + 0.5;
      mean_x += (double)llen_hist[j] * x;
      mean_y += (double)score_sums[j];
      var_x += (double)llen_hist[j] * x * x;
      var_y += (double)score2_sums[j];
      covar_xy += x * (double)score_sums[j];
    }
  mean_x /= n; mean_y /= n;
  var_x = var_x / n - mean_x * mean_x;
  var_y = var_y / n - mean_y * mean_y;
  
  covar_xy = covar_xy / n - mean_x * mean_y;
/*
  rho = covar_xy / var_x;
  mu = mean_y - rho * mean_x;
*/
  mean_y2 = covar_xy2 = var_y2 = 0.0;
  for (j = min_llen; j <= max_llen; j++) 
    if (llen_hist[j] > 1) {
      x = (double)j + 0.5;
      u = rho * x + mu;
      y2 = score2_sums[j] - 2 * score_sums[j] * u + llen_hist[j] * u * u;
      mean_y2 += y2;
      var_y2 += y2 * y2;
      covar_xy2 += x * y2;
      /*      fprintf(stderr,"%6.1lf %4d %8d %8d %7.2lf %8.2lf\n",
	      x,llen_hist[j],score_sums[j],score2_sums[j],u,y2); */
    }
  
  mean_var = mean_y2 /= (double)n;
  covar_xy2 = covar_xy2 / (double)n - mean_x * mean_y2;
  /*
  fprintf(stderr," rho1/mu1: %.4f/%.4f mean_var %.4f\n",
	  rho*LN_FACT,mu,mean_var);
  */
  if (n > 1) var_e = (var_y2/n - mean_y2 * mean_y2)/(n-1);
  else var_e = 0.0;

  if (fit_flag) {
    rho2 = covar_xy2 / var_x;
    mu2 = mean_var - rho2 * mean_x;
  }
  else {
    rho2 = 0;
    mu2 = mean_var;
  }

  if (rho2 < 0.0 )
    z = (rho2 * LN_FACT*log((double)max_length) + mu2 > 0.0) ? max_length : exp((-1.0 - mu2 / rho2)/LN_FACT);
  else z =  rho2 ? exp((1.0 - mu2 / rho2)/LN_FACT) : LENGTH_CUTOFF;
  if (z < 2*LENGTH_CUTOFF) z = 2*LENGTH_CUTOFF;

  var_cutoff = rho2 * LN_FACT*log(z) + mu2;
/*
  fprintf(stderr,"\nMinimum allowed predicted variance (%0.2f) at n = %.0f\n",
	 var_cutoff,z);

  printf("\nVariance of residuals explained by log(n): %.2f\n",
	 covar_xy2 * covar_xy2 / var_x );
 
   printf("\nObserved best fits to log(n):\n");
    printf(" mean = %.2f * log(n) + %.2f,   var = %.2f * log(n) + %.2f \n",
    rho*LN_FACT, mu, rho2*LN_FACT, mu2);

  cell_size = n / 20;

  printf("\n\n                          S              z");
  printf("\n\nLengths (n)   No.     Mean    Std        Mean    Std\n");

  last_ej = LENGTH_CUTOFF;
  for (j = min_llen; j <= max_llen; ) {
    for (s = s2 = t = t2 = u = u2 = n = 0;
	 n < cell_size && j <= max_llen; j++) 
      if (llen_hist[j] > 1) {
	n += llen_hist[j];
	s += score_sums[j];
	s2 += score2_sums[j];
	
	y = j+0.5;
	x = rho * y + mu;
	v = rho2 * y + mu2;
	if (v < var_cutoff) v = var_cutoff;
	u += (score_sums[j] - llen_hist[j] * x) / sqrt(v);
	u2 += (score2_sums[j] - 2 * x * score_sums[j] + llen_hist[j] * x * x) /
	  v;
      }
    if (n > 0) {
      mean = s / (double)n;
      ej = (int)(exp((double)j/LN_FACT)+0.5);
      printf("\n%3d - %4d:   %4d  %6.1f  %6.2f", last_ej, ej, n, mean,
	     sqrt(s2 / (double)n - mean * mean));
      mean = u / n;
      printf("     %6.1f  %6.2f", mean, sqrt(u2 / n - mean * mean));
      last_ej = ej;
    }
  }
*/
}

void
fit_llens()
{
  int j;
  int n, s, s2, ej, last_ej, cell_size, n_u2;
  double mean, x, y, y2, t, t2, u, u2, v, p, z;
  double mean_x, mean_y, var_x, var_y, covar_xy;
  double mean_y2, covar_xy2;
  double mean_u2, mean_3u2, mean_i3u2, dllj;
  double sum_x, sum_y, sum_x2, sum_xy, sum_v, det, n_w;

/* now fit scores to best linear function of log(n), using
   simple linear regression */
  
  for (min_llen=0; min_llen < max_llen; min_llen++)
    if (llen_hist[min_llen]) break;
  min_llen--;

  for (j = min_llen; j < max_llen; j++) {
    if (llen_hist[j] > 1) {
      dllj = (double)llen_hist[j];
      score_var[j] = (double)score2_sums[j]/dllj
	- ((double)score_sums[j]/dllj)
	* ((double)score_sums[j]/dllj);
      score_var[j] /= (double)(llen_hist[j]-1);
      if (score_var[j] <= 1.0 ) score_var[j] = 1.0;
    }
  }
	  
  n_w = 0.0;
  sum_x = sum_y = sum_x2 = sum_xy = sum_v = 0;
  for (j = min_llen; j < max_llen; j++)
    if (llen_hist[j] > 1) {
      x = j + 0.5;
      dllj = (double)llen_hist[j];
      n_w += dllj/score_var[j];
      sum_x +=   dllj * x / score_var[j] ;
      sum_y += (double)score_sums[j] / score_var[j];
      sum_x2 +=  dllj * x * x /score_var[j];
      sum_xy +=  x * (double)score_sums[j]/score_var[j];
    }

  det = n_w * sum_x2 - sum_x * sum_x;
  rho = (n_w * sum_xy  - sum_x * sum_y)/det;
  mu = (sum_x2 * sum_y - sum_x * sum_xy)/det;

/* printf(" rho1/mu1: %.2f/%.2f\n",rho*LN_FACT,mu); */

  n = 0;
  mean_x = mean_y = mean_y2 = 0.0;
  var_x = var_y = 0.0;
  covar_xy = covar_xy2 = 0.0;

  for (j = min_llen; j <= max_llen; j++) 
    if (llen_hist[j] > 1 ) {
      n += llen_hist[j];
      x = (double)j + 0.5;
      dllj = (double)llen_hist[j];
      mean_x += dllj * x;
      mean_y += (double)score_sums[j];
      var_x += dllj * x * x;
      var_y += (double)score2_sums[j];
      covar_xy += x * (double)score_sums[j];
    }
  mean_x /= n; mean_y /= n;
  var_x = var_x / n - mean_x * mean_x;
  var_y = var_y / n - mean_y * mean_y;
  
  covar_xy = covar_xy / n - mean_x * mean_y;
/*  rho = covar_xy / var_x;
  mu = mean_y - rho * mean_x;
*/

  mean_y2 = covar_xy2 = 0.0;
  for (j = min_llen; j <= max_llen; j++) 
    if (llen_hist[j] > 1) {
      x = (double)j + 0.5;
      u = rho * x + mu;
      y2 = score2_sums[j] - 2 * score_sums[j] * u + llen_hist[j] * u * u;
      mean_y2 += y2;
      covar_xy2 += x * y2;
    }
  
  mean_y2 /= n;
  covar_xy2 = covar_xy2 / n - mean_x * mean_y2;
  rho2 = covar_xy2 / var_x;
  mu2 = mean_y2 - rho2 * mean_x;

  if (rho2 < 0.0 )
    z = (rho2 * LN_FACT*log((double)max_length) + mu2 > 0.0) ? max_length : exp((-1.0 - mu2 / rho2)/LN_FACT);
  else z =  rho2 ? exp((1.0 - mu2 / rho2)/LN_FACT) : LENGTH_CUTOFF;
  if (z < 2* LENGTH_CUTOFF) z = 2*LENGTH_CUTOFF;

  var_cutoff = rho2*LN_FACT*log(z) + mu2;

/*  fprintf(stderr,"\nMinimum allowed predicted variance (%0.2f) at n = %.0f\n",
	 var_cutoff,z);
*/
  mean_u2 = 0.0;
  n_u2 = 0;
  for ( j = min_llen; j < max_llen; j++) {
    y = j+0.5;
    dllj = (double)llen_hist[j];
    x = rho * y + mu;
    v = rho2 * y + mu2;
    if (v < var_cutoff) v = var_cutoff;
    if (llen_hist[j]> 1) {
      u2 =  (score2_sums[j] - 2 * x * score_sums[j] + dllj * x * x) - v*dllj;
      mean_u2 += score_var[j] = u2*u2/(llen_hist[j]-1);
      n_u2++;
      /*      fprintf(stderr," %d (%d) u2: %.2f v*ll: %.2f %.2f\n",
	      j,llen_hist[j],u2,v*dllj,sqrt(score_var[j])); */
    }
    else score_var[j] = -1.0;
  }

  mean_u2 = sqrt(mean_u2/(double)n_u2);
  /* fprintf(stderr," mean s.d.: %.2f\n",mean_u2); */

  mean_3u2 = mean_u2*3.0;

  for (j = min_llen; j < max_llen; j++) {
    if (llen_hist[j] <= 1) continue;
    if (sqrt(score_var[j]) > mean_3u2) {
      /*      fprintf(stderr," removing %d %d %.2f\n",
	     j, (int)(exp((double)j/LN_FACT)-0.5),
	     sqrt(score_var[j]));
	     */
      nb_trimmed++;
      n1_trimmed += llen_hist[j];
      llen_hist[j] = 0;
    }
  }
  fit_llen();
}

struct s2str {double s; int n;};
void s2_sort ( struct s2str *sptr, int n);

void
fit_llen2()
{
  int j;
  int n, n_y2, llen_delta, llen_del05, s, s2;
  int n_size;
  double mean, x, y, y2, t, t2, u, u2, v, p, z;
  double mean_x, mean_y, var_x, var_y, covar_xy;
  double mean_y2, covar_xy2, var_y2;
  double ex, exl;
  struct s2str *ss2;
  FILE *tfd;

  double sum_x, sum_y, sum_x2, sum_xy, sum_v, det, n_w;
  
/* now fit scores to best linear function of log(n), using
   simple linear regression */
  
  for (min_llen=0; min_llen < max_llen; min_llen++)
    if (llen_hist[min_llen]) break;

  for ( ; max_llen > min_llen; max_llen--)
    if (llen_hist[max_llen]) break;

  for (n_size=0,j = min_llen; j < max_llen; j++) {
    if (llen_hist[j] > 1) {
      score_var[j] = (double)score2_sums[j]/(double)llen_hist[j]
	- ((double)score_sums[j]/(double)llen_hist[j])
	* ((double)score_sums[j]/(double)llen_hist[j]);
      score_var[j] /= (double)(llen_hist[j]-1);
      if (score_var[j] <= 1.0 ) score_var[j] = 1.0;
      n_size++;
    }
  }
	  
  n_w = 0.0;
  sum_x = sum_y = sum_x2 = sum_xy = sum_v = 0;
  for (j = min_llen; j < max_llen; j++)
    if (llen_hist[j] > 1) {
      x = j + 0.5;
      n_w += (double)llen_hist[j]/score_var[j];
      sum_x +=   (double)llen_hist[j] * x / score_var[j] ;
      sum_y += (double)score_sums[j] / score_var[j];
      sum_x2 +=  (double)llen_hist[j] * x * x /score_var[j];
      sum_xy +=  x * (double)score_sums[j]/score_var[j];
    }

  if (n_size < 5 ) {
    fit_flag=0;
    rho = 0;
    mu = sum_y/n_w;
  }
  else {
    det = n_w * sum_x2 - sum_x * sum_x;
    if (det > 0.001) {
      rho = (n_w * sum_xy  - sum_x * sum_y)/det;
      rho_e = n_w/det;
      mu = (sum_x2 * sum_y - sum_x * sum_xy)/det;
      mu_e = sum_x2/det;
    }
    else {
      fit_flag = 0;
      rho = 0;
      mu = sum_y/n_w;
    }
  }

  det = n_w * sum_x2 - sum_x * sum_x;
  rho = (n_w * sum_xy  - sum_x * sum_y)/det;
  mu = (sum_x2 * sum_y - sum_x * sum_xy)/det;

/*   fprintf(stderr," rho1/mu1: %.2f/%.2f\n",rho*LN_FACT,mu); */

  n = 0;
  mean_x = mean_y = mean_y2 = 0.0;
  var_x = var_y = 0.0;
  covar_xy = covar_xy2 = 0.0;

  for (j = min_llen; j <= max_llen; j++) 
    if (llen_hist[j] > 1 ) {
      n += llen_hist[j];
      x = (double)j + 0.5;
      mean_x += (double)llen_hist[j] * x;
      mean_y += (double)score_sums[j];
      var_x += (double)llen_hist[j] * x * x;
      var_y += (double)score2_sums[j];
      covar_xy += x * (double)score_sums[j];
    }
  mean_x /= n; mean_y /= n;
  var_x = var_x / n - mean_x * mean_x;
  var_y = var_y / n - mean_y * mean_y;
  
  covar_xy = covar_xy / n - mean_x * mean_y;
/*
  rho = covar_xy / var_x;
  mu = mean_y - rho * mean_x;
*/

  if ((ss2=(struct s2str *)calloc(max_llen+1,sizeof(struct s2str)))==NULL) {
    fprintf(stderr," cannot allocate ss2\n");
    return;
  }

  mean_y2 = 0.0;
  n_y2 = n = 0;
  for (j = min_llen; j <= max_llen; j++) 
    if (llen_hist[j] > VHISTC) {
      n++;
      n_y2 += ss2[j].n = llen_hist[j];
      x = (double)j + 0.5;
      u = rho * x + mu;
      ss2[j].s = y2 = score2_sums[j] - 2*score_sums[j]*u + llen_hist[j]*u*u;
      mean_y2 += y2;
    }
  mean_var = mean_y2/(double)n_y2;

  s2_sort(ss2+min_llen,max_llen-min_llen+1);
  
  /*  fprintf(stderr,"min_llen: %d, max: %d\n",min_llen,max_llen); */
  llen_delta = 0;
  for (j=min_llen; j<=max_llen; j++) {
    if (ss2[j].n > 1) {
      llen_delta++;
/*      fprintf(stderr,"%d\t%d\t%.2f\t%.4f\n",
	      j,ss2[j].n,ss2[j].s,ss2[j].s/ss2[j].n);
*/
    }
  }

  llen_del05 = llen_delta/20;
  mean_y2 = 0.0;
  n_y2 = 0;
  for (j = min_llen; j<min_llen+llen_del05; j++) {
    n1_trimmed += ss2[j].n;
    nb_trimmed++;
  }
  for (j = min_llen+llen_del05; j <= min_llen+llen_delta-llen_del05; j++) 
    if (ss2[j].n > 1) {
      mean_y2 += ss2[j].s;
      n_y2 += ss2[j].n;
    }
  for (j = min_llen+llen_delta-llen_del05+1; j< max_llen; j++) {
    n1_trimmed += ss2[j].n;
    nb_trimmed++;
  }
  
  free(ss2);
  if (n_y2 > 1) mean_var = mean_y2/(double)n_y2;

  /*    fprintf(stderr," rho1/mu1: %.4f/%.4f mean_var: %.4f/%d\n",
	  rho*LN_FACT,mu,mean_var,n); */

    var_e = 0.0;
}

float find_z(int score, int length, double comp)
{
  double log_len, var, z;
  
  log_len = LN_FACT*log((double)(length));
  var = rho2 * log_len + mu2;
  if (var < var_cutoff) var = var_cutoff;
  z = ((double)score - rho * log_len - mu) / sqrt(var);

  return (50.0 + 10.0*z);
}

float find_zr(int score, int length, double comp)
{
  double log_len, z;
  
  if ( length < LENGTH_CUTOFF) return 0;

  log_len = LN_FACT*log((double)(length));
/*  var = rho2 * log_len + mu2;
  if (var < var_cutoff) var = var_cutoff;
*/

  z = ((double)score - rho * log_len - mu) / sqrt(mean_var);

  return (50.0 + z*10.0);
}

float find_zr2(int score, int length, double comp)
{
  double log_len, var;
  double z;
  
  if ( length < LENGTH_CUTOFF) return 0;

  log_len = LN_FACT*log((double)(length));

  var = rho2 * log_len + mu2;
  if (var < var_cutoff) var = mean_var;

  z = ((double)score - rho * log_len - mu) / sqrt(var);

  return (50.0 + z*10.0);
}

float find_za(int score, int length, double comp)
{
  double ls, var, z, p_val, mp, np, a_n1, a_n1f, u;
  
  a_n1 = (double)length; 
  a_n1f = log(a_n1)/H;

  mp = a_n0 - a_n0f - a_n1f;
  np = a_n1 - a_n0f - a_n1f;

  if (np < 1.0) np = 1.0;
  if (mp < 1.0) mp = 1.0;

  z = Lambda * score - log(K * np * mp);

  z = -z + 0.577216;
  z /= -1.282555;

  return (50.0 + z*10.0);
}

float find_zn(int score, int length, double comp)
{
  double ls, var;
  float z;
  
  z = ((double)score - mu) / sqrt(mean_var);

  return (50.0 + z*10.0);
}

/* computes E value for a given z value, assuming normal distribution */
double z_to_E(zs)
     double zs;
{

  if (num_db_entries < 5) return (double)num_db_entries;

  if (zs > ZS_MAX) return 0.0;

  return (double)num_db_entries * erfc(zs/M_SQRT2)/2.0; 
}

double zs_to_bit(float zs, int n0, int n1) {
  double e, z;

  /*  if (num_db_entries < 5) return 0.0; */

  z = (zs - 50.0)/10.0;

  return (1.28255 * z + 0.577216 + log((double)(n0*n1)))/LN2 ;
}

/* computes E-value for a given z value, assuming normal distribution */
double zs_to_E(float zs,int n1, int isdna)
{
  double e, z, k;

  if (num_db_entries < 5) return 0.0;

  z = (zs - 50.0)/10.0;

  if (z > ZS_MAX) return 0.0;

  if (isdna == 1) {
    k = (double)total_db.length /(double)n1;
    if (total_db.carry > 0) { k *= (double)total_db.carry * (double)MAXLONG;}
  }
  else k = (double)num_db_entries;

/*  k = (double)num_db_entries;  */

  return k * erfc(z/M_SQRT2)/2.0; 
}

/* computes E-value for a given z value, assuming normal distribution */
float E_to_zs(double E)
{
  double e, z, k;

  e = E/(double)num_db_entries;

  z = (log(e)+0.577216)/(-1.28255);

  return z*10.0+50.0;
}

double zs_to_Ec(float zs) /* computes 1.0 - E value for a given z value,
		    assuming normal distribution */
{
  double z;

  if (num_db_entries < 5) return 0.0;

  z = (zs - 50.0)/10.0;

  if (z > ZS_MAX) return 0.0;

  return (float)num_db_entries*erf(z/M_SQRT2)/2.0; 
}

void
summ_stats(char *s_str)
{
  strcpy(s_str,f_string);

}

void
vsort(v,s,n)
	double *v; int *s, n;
{
  int gap, i, j;
  double tmp;
  int itmp;
	
  for (gap=n/2; gap>0; gap/=2)
    for (i=gap; i<n; i++)
      for (j=i-gap; j>=0; j -= gap) {
	if (v[j] >= v[j+gap]) break;
	tmp = v[j]; v[j]=v[j+gap]; v[j+gap]=tmp;
	itmp = s[j]; s[j]=s[j+gap]; s[j+gap]=itmp;
      }
}

void
calc_ks()
{
  int i, cum_hl;
  double x_tmp, cur_e, f_int, init_e;

  cum_hl = ks_df = 0;
  ks_dev = 0.0;
  f_int = (double)(5*histint + min_hist) + (double)histint/2.0;
  init_e = zs_to_Ec(f_int);
  for (i=0; i<6; i++) cum_hl += hist[i];
  for (i=6; i< (90-min_hist)/histint; i++) {
    cum_hl += hist[i];
    if (hist[i]>0) {
      f_int = (double)(i*histint + min_hist) + (double)histint/2.0;
      cur_e = zs_to_Ec(f_int);
      x_tmp = fabs((double)cum_hl - cur_e);
      if (x_tmp > ks_dev) ks_dev = x_tmp;
      ks_df++;
    }
  }
  ks_dev /= (double)num_db_entries;
}

void s_sort (double **ptr, int nbest)
{
  int gap, i, j;
  double *tmp;

  for (gap = nbest/2; gap > 0; gap /= 2)
    for (i = gap; i < nbest; i++)
      for (j = i - gap; j >= 0; j-= gap) {
	if (*ptr[j] >= *ptr[j + gap]) break;
	tmp = ptr[j];
	ptr[j] = ptr[j + gap];
	ptr[j + gap] = tmp;
      }
}

void st_sort (int *ptr, int n)
{
  int gap, i, j;
  int tmp;

  for (gap = n/2; gap > 0; gap /= 2)
    for (i = gap; i < n; i++)
      for (j = i - gap; j >= 0; j-= gap) {
	if (ptr[j] >= ptr[j + gap]) break;
	tmp = ptr[j];
	ptr[j] = ptr[j + gap];
	ptr[j + gap] = tmp;
      }
}


void s2_sort (struct s2str *ptr, int n)
{
  int gap, i, j;
  struct s2str tmp;

  for (gap = n/2; gap > 0; gap /= 2)
    for (i = gap; i < n; i++)
      for (j = i - gap; j >= 0; j-= gap) {
	if (ptr[j].s >= ptr[j + gap].s) break;
	tmp.s = ptr[j].s;
	tmp.n = ptr[j].n;
	ptr[j].s = ptr[j + gap].s;
	ptr[j].n = ptr[j + gap].n;
	ptr[j + gap].s = tmp.s;
	ptr[j + gap].n = tmp.n;
      }
}
